/*
 * Code Mercenaries Hard- und Software GmbH
 * Karl-Marx-Strasse 147A
 * 12529 Schönefeld
 * Deutschland
 * 
 * August 2015
 * 
 * Used result values:
 *   -1 hard error
 *       1 light weight error (like wrong user input)
 *    0 success
 * 
 * This program is using the BCM2835 library for 
 * Raspberry Pi 2. More information at
 * http://www.airspayce.com/mikem/bcm2835/
 * */

//register
#define LW04_REG_NULL      0x00
#define LW04_REG_INFO      0x01

#define LW04_REG_CURRENT_1 0x02
#define LW04_REG_CURRENT_2 0x03
#define LW04_REG_CURRENT_3 0x04
#define LW04_REG_CURRENT_4 0x05

#define LW04_REG_OUTPUT_1  0x06
#define LW04_REG_OUTPUT_2  0x07
#define LW04_REG_OUTPUT_3  0x08
#define LW04_REG_OUTPUT_4  0x09

#define LW04_REG_STATUS    0x0A
#define LW04_REG_MODE_0    0x0B
#define LW04_REG_MODE_1    0x0C
#define LW04_REG_DMX       0x0D
#define LW04_REG_FLASH     0x0E
#define LW04_REG_SEQ       0x0F

//write/ read mode
#define LW04_WRITE 0
#define LW04_READ  1

//include read and write methods with error check
#include "errorCheckedBCM2835.h"

void  LW04_setSlave(int);
char  LW04_setStartregister(int);
char  LW04_listDevices();
char  LW04_saveDataPermanent(char value);

char  LW04_getColorOutput(int, int*);
char  LW04_setColorOutput(int, int);

char  LW04_getMode(int, char*);
char  LW04_setMode(char reg, char value);

char  LW04_getCurrent(int, int*);
char  LW04_setCurrent(char, int);

short LW04_convertColor(int);
int   LW04_negConvertColor(short, short);

char  LW04_enableChannel(char);
char  LW04_disableChannel(char);

char  LW04_testDevice();
char  LW04_readInfo(char*, char*, char*);
char  LW04_dimmer(char mode);


/*
 * Method to write a value in to the register of the device
 * */
char LW04_setColorOutput(int reg, int value) {
    
    char buffer[3];
    buffer[0] = reg;
    buffer[1] = value & 0x00FF;
    buffer[2] = (value & 0xFF00) >> 8;
    
    //write 3 bytes
    char result = errorCheck_bcm2835_i2c_write(buffer, 3);
    if(result != 0) 
		return -1;
    
    return 0;
}

/*
 * Method to read data from the register of the device
 * */
char LW04_getColorOutput(int reg, int* value) {
    
    char buffer[2];
    char result;
    
    //set pointer to reg for reading task
    result = LW04_setStartregister(reg);
    if(result != 0) 
		return -1;
    
    //read 2 bytes in to buffer
    result = errorCheck_bcm2835_i2c_read(buffer, 2);
    if(result != 0) 
		return -1;
    
    *value = LW04_negConvertColor(buffer[0], buffer[1]);
    
    return 0;
}

/*
 * Set intern device pointer to reg
 * */
char LW04_setStartregister(int reg) {
    
    char buffer[1];//only set register, not write data
    buffer[0] = reg;
	
    //write one byte with the destination register
    char result = errorCheck_bcm2835_i2c_write(buffer, 1);
    if(result != 0) 
		return -1;
    
    return 0;
}

/*
 * convert input (given in percentages (0-100))
 * */
short LW04_convertColor(int input) {
    
    return input * 40.95;
}

/*
 * reconvert (inverted convertColor())
 * */
int LW04_negConvertColor(short a, short b) {
    
    int temp = ((b << 8) & 0x0F00) | a;
    int result = temp / 40.95;
    
    return result;
}

/*
 * Set the address we are reading from and
 * writing to
 * */
void LW04_setSlave(int address) {
    
    bcm2835_i2c_setSlaveAddress(address);
}

/*
 * start test mode
 * */
char LW04_testDevice() {
    
    char buffer[2];
    char mode;
    char result = LW04_getMode(LW04_REG_MODE_1, &mode);
    if(result != 0) 
		return -1;
    
    buffer[0] = LW04_REG_MODE_1;
    buffer[1] = 0x80 | mode;
    
    result = errorCheck_bcm2835_i2c_write(buffer, 2);
    if(result != 0) 
		return -1;
    
    return 0;
}

/*
 * get Infos about Led Warrior04
 * */
char LW04_readInfo(char* serialNumber, char* version, char* model) {
    
    char result;
   
    //set pointer to reg for reading task
    result = LW04_setStartregister(LW04_REG_INFO);
    if(result != 0) 
		return -1;
    
    char buffer[9] = {0};
	
    //read 9 bytes wich contain the information
    result = errorCheck_bcm2835_i2c_read(buffer, 9);
    if(result != 0) 
		return -1;
    
    sprintf(serialNumber, "Serial Number: %02x%02x%02x%02x", buffer[4], buffer[3], buffer[2], buffer[1]);
    sprintf(version, "Version: %02x%02x", buffer[6], buffer[5]);
    sprintf(model, "Model: LED-Warrior %02x%02x", buffer[8], buffer[7]);
    
    return 0;
}

/*
 * Set current on specified channel
 * */
char LW04_setCurrent(char reg, int current) {
    
    char mode;
    char result = LW04_getMode(LW04_REG_MODE_0, &mode);
    if(result != 0) 
		return -1;
    
    char res = mode & 0x7F;
    
	//Open current register
    result = LW04_setMode(LW04_REG_MODE_0, res | 0x80);
    if(result != 0) 
		return -1;
    
    char buffer[2];
    buffer[0] = reg;
    buffer[1] = current / 5; //divide through 5 because 1 bit = 5mA
    
    //set new current
    result = errorCheck_bcm2835_i2c_write(buffer, 2);
    if(result != 0) 
		return -1;
    
    return 0;
}

/*
 * Get current from specified channel
 * */
char LW04_getCurrent(int reg, int* value) {
    
    char buffer[1];
    char result;
    
    //set pointer to reg for reading task
    result = LW04_setStartregister(reg);
    if(result != 0) 
		return -1;
    
    result = errorCheck_bcm2835_i2c_read(buffer, 1);
    if(result != 0) 
		return -1;
    
    *value = buffer[0] * 5; //multiple with 5 because 1 bit = 5mA
    
    return 0;
}

/*
 * set dimmer mode by number
 * */
char LW04_dimmer(char input_mode) {
    
    char mode = 0;
	char modeRes;
	
    switch(input_mode) {
        
        case 0:     mode = 0x00; //I2C
                    break;
                
        case 1:     mode = 0x10; //DMX
                    break;
        
        case 2:     mode = 0x20; //DALI
                    break;
        
        case 3:     mode = 0x30; //SEQUENCE
                    break;
                
        default:    mode = 255;
    }
    
    if(mode == 255) {
        printf("Incorrect mode %d!\n", input_mode);
        return 1;
    }
    
    char result = LW04_getMode(LW04_REG_MODE_0, &modeRes);
	if(result == -1) 
		return -1;
    
    result = LW04_setMode(LW04_REG_MODE_0, mode | (modeRes & 0xCF));
	if(result == -1) 
		return -1;
    
    return 0;
}

/*
 * Disable channel by channelnumber
 * */
char LW04_disableChannel(char channel) {
    
    char chan = 0;
	char mode;
    switch(channel) {
        
        case 0:     chan = 0xFE;
                    break;
                
        case 1:     chan = 0xFD;
                    break;
        
        case 2:     chan = 0xFB;
                    break;
        
        case 3:     chan = 0xF7;
                    break;
                
        default:    chan = 255;
    }
    
    if(chan == 255) {
        printf("Incorrect channel %d!\n", channel);
        return 1;
    }
    
    char result = LW04_getMode(LW04_REG_MODE_0, &mode);
    if(result == -1) 
		return -1;
    
    result = LW04_setMode(LW04_REG_MODE_0, chan & mode);
    if(result == -1) 
		return -1;
    
    return 0;
}

/*
 * Enable channel by channelnumber
 * */
char LW04_enableChannel(char channel) {
    
    char chan = 0;
	char mode;
	
    switch(channel) {
        
        case 0:     chan = 0x01;
                    break;
                
        case 1:     chan = 0x02;
                    break;
        
        case 2:     chan = 0x04;
                    break;
        
        case 3:     chan = 0x08;
                    break;
                
        default:    chan = 255;
    }
    
    if(chan == 255) {
        printf("Incorrect channel %d!\n", channel);
        return 1;
    }
    
    char result = LW04_getMode(LW04_REG_MODE_0, &mode);
    if(result == -1) 
		return -1;
    
    result = LW04_setMode(LW04_REG_MODE_0, chan | mode);
    if(result == -1) 
		return -1;
    
    return 0;
}

/*
 * Get mode register
 * */
char LW04_getMode(int reg, char* dest) {
    
    char result;
    char buffer[1] = {0};
	
    result = LW04_setStartregister(reg);
    if(result != 0) 
		return -1;  
    
    result = errorCheck_bcm2835_i2c_read(buffer, 1);
    if(result != 0) 
		return -1;
    
    *dest = buffer[0];
    
    return 0;
}

/*
 * Set mode register
 * */
char LW04_setMode(char reg, char value) {
    
    char buffer[2] = {0};
    buffer[0] = reg;
    buffer[1] = value;
    
    char result = errorCheck_bcm2835_i2c_write(buffer, 2);
    if(result != 0) 
		return -1;
    
    return 0;
}

/*
 * Save specified data to be saved permanent on chip
 * */
char LW04_saveDataPermanent(char value) {
    
    char buffer[2] = {0};
    buffer[0] = LW04_REG_FLASH;
    buffer[1] = (value<<4) & 0x70;
    
    char result = errorCheck_bcm2835_i2c_write(buffer, 2);
    if(result != 0) 
		return -1;
    
    return 0;
}

/*
 * Get a list with all active devices on I2C bus
 * */
char LW04_listDevices() {
    
    //in this method we are using the standard read and write method,
    //so we dont get a possible NACK for every address
    
    char bufferRead[9] = {0};
    
    //scan through all device addresses (0 is broadcast)
    for(int i = 1; i<127; i++) {
        
        //clear buffer
        memset(bufferRead, 0, 9);
        
        //change address
        LW04_setSlave(i);
        
        //set pointer to reg for reading task
        char bufferWrite[1];//only set register, not write data
        bufferRead[0] = LW04_REG_NULL;
		
        //write one byte with the destination register
        bcm2835_i2c_write(bufferWrite, 1);        
        
        bcm2835_i2c_read(bufferRead, 9);
        if(bufferRead[0] >= 1) {
            printf("Device found on address %d\r\n", i);
        }
    }
    
    return 0;
}
